from abc import ABC, abstractmethod
import pathlib
from typing import Any, Dict, Optional
import uuid

import requests

from anyscale.client.openapi_client.api.default_api import DefaultApi as InternalApi
from anyscale.client.openapi_client.models import (
    CloudDataBucketFileType,
    CloudDataBucketPresignedUploadInfo,
    CloudDataBucketPresignedUploadRequest,
    Project,
)
from anyscale.controllers.service_controller import ServiceController
from anyscale.utils.cloud_utils import get_default_cloud
from anyscale.utils.runtime_env import zip_local_dir


RUNTIME_ENV_PACKAGE_FORMAT = "pkg_{content_hash}.zip"


class AnyscaleClientWrapper(ABC):
    @abstractmethod
    def get_default_project_id(self) -> str:
        """Get the default project ID for this user."""
        raise NotImplementedError

    @abstractmethod
    def get_cloud_id(self, cloud_name: Optional[str] = None) -> str:
        """Get the cloud ID for the provided cloud name.

        If cloud name is None, gets the default cloud for this user.
        """
        raise NotImplementedError

    @abstractmethod
    def roll_out_service(self, config_dict: Dict[str, Any]):
        """Deploy or update the service to use the provided config."""
        # NOTE(edoakes): using the untyped dictionary is temporary and this will be updated
        # to directly call the SDK with the ApplyServiceModel.
        raise NotImplementedError

    @abstractmethod
    def upload_local_dir_to_cloud_storage(
        self, local_dir: str, *, cloud_id: str
    ) -> str:
        """Upload the provided directory to cloud storage and return a URI for it.

        The directory will be zipped and the resulting URI can be in a Ray runtime_env.

        The upload is preformed using a pre-signed URL fetched from Anyscale, so no
        local cloud provider authentication is required.
        """
        raise NotImplementedError


class FakeAnyscaleClient(AnyscaleClientWrapper):
    CLOUD_BUCKET = "s3://fake-bucket/{cloud_id}"
    DEFAULT_CLOUD_ID = "fake-default-cloud-id"
    DEFAULT_PROJECT_ID = "fake-default-project-id"

    def __init__(self):
        self._rolled_out_config: Optional[Dict[str, Any]] = None

    def get_default_project_id(self) -> str:
        return self.DEFAULT_PROJECT_ID

    def get_cloud_id(self, cloud_name: Optional[str] = None) -> str:
        if cloud_name is None:
            return self.DEFAULT_CLOUD_ID

        return f"{cloud_name}-fake-id"

    @property
    def rolled_out_service(self) -> Optional[Dict[str, Any]]:
        return self._rolled_out_config

    def roll_out_service(self, config_dict: Dict[str, Any]):
        self._rolled_out_config = config_dict

    def upload_local_dir_to_cloud_storage(
        self, local_dir: str, *, cloud_id: str  # noqa: ARG002
    ) -> str:
        bucket = self.CLOUD_BUCKET.format(cloud_id=cloud_id)
        return f"{bucket}/fake_pkg_{uuid.uuid4()}.zip"


class RealAnyscaleClient(AnyscaleClientWrapper):
    def __init__(
        self,
        *,
        controller: Optional[ServiceController] = None,
        internal_api_client: Optional[InternalApi] = None,
    ):
        # NOTE(edoakes): this is an incremental path, eventually the logic in the
        # service controller will be moved here.
        self._controller = controller or ServiceController()
        self._internal_api_client = (
            internal_api_client or self._controller.internal_api_client
        )

        # Cached IDs to avoid duplicate lookups.
        self._default_project_id: Optional[str] = None
        self._cloud_id_cache: Dict[Optional[str], str] = {}

    def get_default_project_id(self) -> str:
        if self._default_project_id is None:
            default_project: Project = self._internal_api_client.get_default_project_api_v2_projects_default_project_get().result
            self._default_project_id = default_project.id

        return self._default_project_id

    def get_cloud_id(self, cloud_name: Optional[str] = None) -> str:
        if cloud_name in self._cloud_id_cache:
            return self._cloud_id_cache[cloud_name]

        # TODO(edoakes): this fetch path requires multiple RTTs which seems unnecessary.
        # We should consider adding a better backend endpoint to hit.
        cloud_id, _ = get_default_cloud(self._internal_api_client, cloud_name)
        self._cloud_id_cache[cloud_name] = cloud_id
        return cloud_id

    def roll_out_service(self, config_dict: Dict[str, Any]):
        self._controller.rollout(config_dict, auto_complete_rollout=True)

    def upload_local_dir_to_cloud_storage(
        self, local_dir: str, *, cloud_id: str
    ) -> str:
        if not pathlib.Path(local_dir).is_dir():
            raise RuntimeError(f"Path '{local_dir}' is not a valid directory.")

        with zip_local_dir(local_dir) as (_, zip_file_bytes, content_hash):
            request = CloudDataBucketPresignedUploadRequest(
                file_type=CloudDataBucketFileType.RUNTIME_ENV_PACKAGES,
                file_name=RUNTIME_ENV_PACKAGE_FORMAT.format(content_hash=content_hash),
            )
            info: CloudDataBucketPresignedUploadInfo = self._internal_api_client.generate_cloud_data_bucket_presigned_upload_url_api_v2_clouds_cloud_id_generate_cloud_data_bucket_presigned_upload_url_post(
                cloud_id, request
            ).result
            requests.put(info.upload_url, data=zip_file_bytes).raise_for_status()

        return info.file_uri



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import base64
import re
from typing import Any, Dict
import unicodedata


def execution_log_name(session_command_id: str) -> str:
    # Note: this path is no longer correct. We should remove this once
    # we've cleaned up the logic backing `/api/v2/session_commands/{session_command_id}/execution_logs`
    # and migrate the webterminal to a sidecar/standalone executable.
    return "/tmp/ray_command_output_{session_command_id}".format(
        session_command_id=session_command_id
    )


def startup_log_name(session_id: str) -> str:
    return f"/tmp/session_startup_logs_{session_id}"


def startup_log_name_v2(session_id: str, encoded: bool = False) -> str:
    startup_log_name = "/var/log/anyscale/session_startup_logs_{session_id}".format(
        session_id=session_id
    )

    if encoded:
        # The encoding should be the same as `encodePathToFileName` in go/infra/anyscaled/internal/logcache/utils.go
        return base64.standard_b64encode(f"{startup_log_name}.out".encode()).decode()
    else:
        return startup_log_name


def slugify(value: str) -> str:
    """
    Code adopted from here https://github.com/django/django/blob/master/django/utils/text.py

    Convert  to ASCII. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Also strip leading and trailing whitespace.
    """

    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value).strip()
    return re.sub(r"[-\s]+", "-", value)


def get_container_name(cluster_config: Dict[str, Any]) -> str:
    return str(cluster_config.get("docker", {}).get("container_name", ""))



import copy
from typing import Any, Dict, List, Optional, Union

import yaml

from anyscale.cli_logger import BlockLogger
from anyscale.service._private.anyscale_client import AnyscaleClientWrapper
from anyscale.utils.runtime_env import is_dir_remote_uri, parse_requirements_file


logger = BlockLogger()


class ServiceConfig:
    def __init__(
        self,
        *,
        import_path: Optional[str] = None,
        arguments: Optional[Dict[str, Any]] = None,
        ray_serve_config: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        image: Optional[str] = None,
        working_dir: Optional[str] = None,
        requirements: Optional[Union[str, List[str]]] = None,
        **kwargs,
    ):
        if len(kwargs) > 0:
            raise ValueError(f"Unrecognized options: {list(kwargs.keys())}.")

        if name is not None and not isinstance(name, str):
            raise TypeError("name must be a string.")

        if image is not None and not isinstance(image, str):
            raise TypeError("image must be a string.")

        self._name = name
        self._image = image
        self._ray_serve_config = self._populate_ray_serve_config(
            import_path=import_path,
            arguments=arguments,
            ray_serve_config=ray_serve_config,
            working_dir=working_dir,
            requirements=self._parse_requirements(requirements),
        )

    def _populate_ray_serve_config(
        self,
        import_path: Optional[str] = None,
        arguments: Optional[Dict[str, str]] = None,
        ray_serve_config: Optional[Dict[str, Any]] = None,
        working_dir: Optional[str] = None,
        requirements: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        # Generate a minimal ray_serve_config from the import_path and arguments if needed.
        if ray_serve_config is not None:
            if import_path is not None or arguments is not None:
                raise ValueError(
                    "import_path and arguments cannot be provided with ray_serve_config."
                )

            # Copy to avoid modifying arguments passed directly from user code.
            ray_serve_config = copy.deepcopy(ray_serve_config)
        elif import_path is not None:
            ray_serve_config = {
                "applications": [{"import_path": import_path,}],
            }
            if arguments is not None:
                if not isinstance(arguments, dict):
                    raise TypeError("Application arguments must be a dictionary.")
                ray_serve_config["applications"][0]["arguments"] = arguments
        else:
            raise ValueError("Either import_path or ray_serve_config must be provided.")

        applications = ray_serve_config.get("applications", None)
        if not applications:
            raise ValueError("ray_serve_config must contain a list of applications.")

        ray_serve_config["applications"] = self._override_application_runtime_envs(
            applications, working_dir=working_dir, requirements=requirements,
        )

        return ray_serve_config

    def _parse_requirements(
        self, requirements: Optional[Union[str, List[str]]] = None
    ) -> Optional[List[str]]:
        parsed_requirements = None
        if isinstance(requirements, str):
            parsed_requirements = parse_requirements_file(requirements)
            if parsed_requirements is None:
                raise ValueError(f"Requirements file {requirements} does not exist.")
        elif isinstance(requirements, list):
            # Copy to avoid modifying arguments passed directly from user code.
            parsed_requirements = copy.deepcopy(requirements)

        return parsed_requirements

    def _override_application_runtime_envs(
        self,
        applications: List[Dict[str, Any]],
        *,
        working_dir: Optional[str],
        requirements: Optional[List[str]],
    ):
        """Override the runtime_env field of the provided applications.

        Only the working_dir and requirements fields will be modified.
        """
        applications = copy.deepcopy(applications)
        for application in applications:
            runtime_env = application.get("runtime_env", {})
            if working_dir is not None:
                runtime_env["working_dir"] = working_dir

            if requirements is not None:
                runtime_env["pip"] = requirements

            if runtime_env:
                application["runtime_env"] = runtime_env

        return applications

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        return cls(**d)

    @classmethod
    def from_yaml(cls, path: str):
        with open(path) as f:
            return cls.from_dict(yaml.safe_load(f))

    @property
    def name(self) -> Optional[str]:
        # TODO(edoakes): in the future, this should get populated from workspace
        # or default immediately and never be None here.
        return self._name

    @property
    def image(self) -> Optional[str]:
        # TODO(edoakes): in the future, this should get populated from workspace
        # or default immediately and never be None here.
        return self._image

    @property
    def ray_serve_config(self) -> Dict[str, Any]:
        return self._ray_serve_config

    def options(
        self,
        ray_serve_config: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        image: Optional[str] = None,
        working_dir: Optional[str] = None,
        requirements: Optional[Union[str, List[str]]] = None,
    ) -> "ServiceConfig":
        return ServiceConfig(
            ray_serve_config=ray_serve_config or self.ray_serve_config,
            name=name or self.name,
            image=image or self.image,
            working_dir=working_dir,
            requirements=requirements,
        )

    def with_local_dirs_uploaded(
        self, client: AnyscaleClientWrapper
    ) -> "ServiceConfig":
        """Returns a copy of the config with all local dirs converted to remote URIs.

        Local dirs can be specified in the working_dir or py_modules fields of the runtime_env.

        Each unique local directory across these fields will be uploaded once to cloud storage,
        then all occurrences of it in the config will be replaced with the corresponding remote URI.
        """
        cloud_id = client.get_cloud_id()
        new_ray_serve_config = copy.deepcopy(self.ray_serve_config)

        local_path_to_uri: Dict[str, str] = {}

        def _upload_dir_memoized(target: str) -> str:
            if is_dir_remote_uri(target):
                return target
            if target in local_path_to_uri:
                return local_path_to_uri[target]

            logger.info(f"Uploading local dir '{target}' to cloud storage.")
            uri = client.upload_local_dir_to_cloud_storage(target, cloud_id=cloud_id)
            local_path_to_uri[target] = uri
            return uri

        for application in new_ray_serve_config["applications"]:
            working_dir = application.get("runtime_env", {}).get("working_dir", None)
            if working_dir is not None:
                application["runtime_env"]["working_dir"] = _upload_dir_memoized(
                    working_dir
                )

            py_modules = application.get("runtime_env", {}).get("py_modules", None)
            if py_modules is not None:
                new_py_modules = []
                for py_module in py_modules:
                    new_py_modules.append(_upload_dir_memoized(py_module))

                application["runtime_env"]["py_modules"] = new_py_modules

        return self.options(ray_serve_config=new_ray_serve_config)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ServiceConfig):
            return all(
                [
                    self.ray_serve_config == other.ray_serve_config,
                    self.image == other.image,
                    self.name == other.name,
                ]
            )

        return False


class ServiceStatus:
    # TODO: implement __str__ and __repr__.
    def __init__(self):
        pass

    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class GCPFileStoreConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'instance_name': 'str',
        'mount_target_ip': 'str',
        'root_dir': 'str'
    }

    attribute_map = {
        'instance_name': 'instance_name',
        'mount_target_ip': 'mount_target_ip',
        'root_dir': 'root_dir'
    }

    def __init__(self, instance_name=None, mount_target_ip=None, root_dir=None, local_vars_configuration=None):  # noqa: E501
        """GCPFileStoreConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._instance_name = None
        self._mount_target_ip = None
        self._root_dir = None
        self.discriminator = None

        self.instance_name = instance_name
        self.mount_target_ip = mount_target_ip
        self.root_dir = root_dir

    @property
    def instance_name(self):
        """Gets the instance_name of this GCPFileStoreConfig.  # noqa: E501


        :return: The instance_name of this GCPFileStoreConfig.  # noqa: E501
        :rtype: str
        """
        return self._instance_name

    @instance_name.setter
    def instance_name(self, instance_name):
        """Sets the instance_name of this GCPFileStoreConfig.


        :param instance_name: The instance_name of this GCPFileStoreConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and instance_name is None:  # noqa: E501
            raise ValueError("Invalid value for `instance_name`, must not be `None`")  # noqa: E501

        self._instance_name = instance_name

    @property
    def mount_target_ip(self):
        """Gets the mount_target_ip of this GCPFileStoreConfig.  # noqa: E501


        :return: The mount_target_ip of this GCPFileStoreConfig.  # noqa: E501
        :rtype: str
        """
        return self._mount_target_ip

    @mount_target_ip.setter
    def mount_target_ip(self, mount_target_ip):
        """Sets the mount_target_ip of this GCPFileStoreConfig.


        :param mount_target_ip: The mount_target_ip of this GCPFileStoreConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and mount_target_ip is None:  # noqa: E501
            raise ValueError("Invalid value for `mount_target_ip`, must not be `None`")  # noqa: E501

        self._mount_target_ip = mount_target_ip

    @property
    def root_dir(self):
        """Gets the root_dir of this GCPFileStoreConfig.  # noqa: E501


        :return: The root_dir of this GCPFileStoreConfig.  # noqa: E501
        :rtype: str
        """
        return self._root_dir

    @root_dir.setter
    def root_dir(self, root_dir):
        """Sets the root_dir of this GCPFileStoreConfig.


        :param root_dir: The root_dir of this GCPFileStoreConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and root_dir is None:  # noqa: E501
            raise ValueError("Invalid value for `root_dir`, must not be `None`")  # noqa: E501

        self._root_dir = root_dir

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GCPFileStoreConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GCPFileStoreConfig):
            return True

        return self.to_dict() != other.to_dict()



from json import dumps as json_dumps
import pathlib
from typing import Dict, Optional, Tuple

import click
from typing_extensions import Literal

from anyscale.cli_logger import BlockLogger
from anyscale.controllers.service_controller import ServiceController
import anyscale.service
from anyscale.service import ServiceConfig, ServiceStatus
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group("service")
def service_cli():
    pass


def _convert_arguments_to_dict(args: Tuple[str]) -> Dict[str, str]:
    """Convert args of the form "key=val" into a dictionary of {key: val}.

    NOTE(edoakes): this mimics the functionality of the `serve run` CLI and should be
    kept in sync with it.
    """
    args_dict = {}
    for arg in args:
        split = arg.split("=")
        if len(split) != 2:
            raise click.ClickException(
                f"Invalid application argument '{arg}', "
                "must be of the form '<key>=<val>'."
            )

        args_dict[split[0]] = split[1]

    return args_dict


@service_cli.command(
    name="deploy",
    short_help="Deploy or update a service.",
    help=(
        "Deploy or update a service.\n\n"
        "If no service with the provided name is running, one will be created, "
        "else the existing service will be updated.\n\n"
        "The target of this command can either be an import path to a Ray Serve "
        "application or a path to a service config file.\n\n"
        "Command-line flags take precedence over values in the config file."
    ),
    hidden=True,
)
@click.argument("import_path_or_config", type=str)
@click.argument("arguments", nargs=-1, required=False)
@click.option(
    "-n",
    "--name",
    required=False,
    default=None,
    type=str,
    help="Unique name for the service.",
)
@click.option(
    "--image",
    required=False,
    default=None,
    type=str,
    help="Container image to use for the service. This cannot be changed when using --in-place. When running in a workspace, this defaults to the image of the workspace.",
)
@click.option(
    "--working-dir",
    required=False,
    default=None,
    type=str,
    help="Path to a local directory that will be the working directory for the service. The files in the directory will be automatically uploaded to cloud storage. When running in a workspace, this defaults to the current working directory.",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the service. These will be installed on top of the image. When running in a workspace, this defaults to the workspace dependencies.",
)
@click.option(
    "-i",
    "--in-place",
    is_flag=True,
    show_default=True,
    default=False,
    help="Perform an in-place upgrade without starting a new cluster. This can be used for faster iteration during development but is *not* currently recommended for production deploys.",
)
@click.option(
    "--canary-percent",
    required=False,
    default=None,
    type=int,
    help="The percentage of traffic to send to the canary version of the service (0-100). This can be used to manually shift traffic toward (or away from) the canary version. If not provided, traffic will be shifted incrementally toward the canary version until it reaches 100. Not supported when using --in-place.",
)
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Amount of excess capacity allowed to be used while updating the service (0-100). Defaults to 100. Not supported when using --in-place.",
)
def deploy(  # noqa: PLR0913
    import_path_or_config: str,
    arguments: Tuple[str],
    name: Optional[str],
    image: Optional[str],
    working_dir: Optional[str],
    requirements: Optional[str],
    in_place: bool,
    canary_percent: Optional[int],
    max_surge_percent: Optional[int],
):
    if import_path_or_config.endswith(".yaml"):
        if not pathlib.Path(import_path_or_config).is_file():
            raise click.ClickException(
                f"Config file '{import_path_or_config}' not found."
            )

        if len(arguments) > 0:
            raise click.ClickException(
                "Application arguments can't be passed when deploying from a config file. "
                "Include them in the config file instead."
            )

        config = ServiceConfig.from_yaml(import_path_or_config)

    else:
        config = ServiceConfig(
            import_path=import_path_or_config,
            arguments=_convert_arguments_to_dict(arguments) or None,
        )

    anyscale.service.deploy(
        config.options(
            name=name, image=image, working_dir=working_dir, requirements=requirements,
        ),
        in_place=in_place,
        canary_percent=canary_percent,
        max_surge_percent=max_surge_percent,
    )


@service_cli.command(
    name="status", help="Get the status of a service.", hidden=True,
)
@click.argument("name", type=str)
@click.option(
    "--json",
    "-j",
    is_flag=True,
    show_default=True,
    default=False,
    help="Output the status in a structured JSON format.",
)
def status(name: str, json: bool):
    status: ServiceStatus = anyscale.service.status(name)
    if json:
        print(json_dumps(status.to_dict()))
    else:
        print(status)


@service_cli.command(
    name="rollout", help="Roll out a service.",
)
@click.option(
    "-f",
    "--service-config-file",
    required=True,
    help="The path of the service configuration file.",
)
@click.option("--name", "-n", required=False, default=None, help="Name of service.")
@click.option("--version", required=False, default=None, help="Version of service.")
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Max amount of excess capacity allocated during the rollout (0-100).",
)
@click.option(
    "--canary-percent",
    required=False,
    default=None,
    type=int,
    help="The percentage of traffic to send to the canary version of the service.",
)
@click.option(
    "--rollout-strategy",
    required=False,
    default=None,
    type=click.Choice(["ROLLOUT", "IN_PLACE"], case_sensitive=False),
    help="Strategy for rollout.",
)
@click.option(
    "-i",
    "--in-place",
    "in_place",
    is_flag=True,
    show_default=True,
    default=False,
    help="Alias for `--rollout-strategy=IN_PLACE`.",
)
@click.option(
    "--no-auto-complete-rollout",
    is_flag=True,
    show_default=True,
    default=False,
    help="Do not complete the rollout (terminate the existing version cluster) after the canary version reaches 100%",
)
def rollout(  # noqa: PLR0913
    service_config_file: str,
    name: Optional[str],
    version: Optional[str],
    max_surge_percent: Optional[int],
    canary_percent: Optional[int],
    rollout_strategy: Optional[Literal["ROLLOUT", "IN_PLACE"]],
    in_place: bool,
    no_auto_complete_rollout: bool,
):
    """Start or update a service rollout to a new version."""
    if in_place:
        if rollout_strategy is not None:
            raise click.ClickException(
                "Only one of `--in-place/-i` and `--rollout-strategy` can be provided."
            )
        rollout_strategy = "IN_PLACE"

    service_controller = ServiceController()
    service_config = service_controller.read_service_config_file(service_config_file)
    service_controller.rollout(
        service_config,
        name=name,
        version=version,
        max_surge_percent=max_surge_percent,
        canary_percent=canary_percent,
        rollout_strategy=rollout_strategy,
        auto_complete_rollout=not no_auto_complete_rollout,
    )


@service_cli.command(name="list", help="Display information about services.")
@click.option(
    "--name", "-n", required=False, default=None, help="Filter by service name."
)
@click.option(
    "--service-id", "--id", required=False, default=None, help="Filter by service id."
)
@click.option(
    "--project-id", required=False, default=None, help="Filter by project id."
)
@click.option(
    "--created-by-me",
    help="List services created by me only.",
    is_flag=True,
    default=False,
)
@click.option(
    "--max-items",
    required=False,
    default=10,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list(  # noqa: A001
    name: Optional[str],
    service_id: Optional[str],
    project_id: Optional[str],
    created_by_me: bool,
    max_items: int,
):
    """List services based on the provided filters.

    This returns both v1 and v2 services.
    """
    service_controller = ServiceController()
    service_controller.list(
        name=name,
        service_id=service_id,
        project_id=project_id,
        created_by_me=created_by_me,
        max_items=max_items,
    )


@service_cli.command(name="rollback", help="Roll back a service.")
@click.option(
    "--service-id", "--id", default=None, help="ID of service.",
)
@click.option("--name", "-n", required=False, default=None, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "--service-config-file", "-f", help="The path of the service configuration file.",
)
@click.option(
    "--max-surge-percent",
    required=False,
    default=None,
    type=int,
    help="Max amount of excess capacity allocated during the rollback (0-100).",
)
def rollback(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    service_config_file: Optional[str],
    max_surge_percent: Optional[int],
):
    """Perform a rollback for a service that is currently in a rollout.

    This *only* applies to v2 services.
    """
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=service_config_file,
        project_id=project_id,
    )
    service_controller.rollback(service_id, max_surge_percent)


@service_cli.command(name="terminate", help="Terminate a service.")
@click.option(
    "--service-id", "--id", required=False, help="ID of service.",
)
@click.option("--name", "-n", required=False, help="Name of service.")
@click.option("--project-id", required=False, help="Filter by project id.")
@click.option(
    "--service-config-file", "-f", help="The path of the service configuration file.",
)
def terminate(
    service_id: Optional[str],
    name: Optional[str],
    project_id: Optional[str],
    service_config_file: Optional[str],
):
    """Terminate a service.

    This applies to both v1 and v2 services.
    """
    service_controller = ServiceController()
    service_id = service_controller.get_service_id(
        service_id=service_id,
        service_name=name,
        service_config_file=service_config_file,
        project_id=project_id,
    )
    service_controller.terminate(service_id)



from functools import wraps
from typing import Callable, Optional

from anyscale.service._private.sdk import ServiceSDK
from anyscale.service.models import ServiceConfig, ServiceStatus


_LAZY_GLOBAL_SDK: Optional[ServiceSDK] = None


def _inject_global_sdk(f: Callable):
    @wraps(f)
    def wrapper(*args, **kwargs):
        global _LAZY_GLOBAL_SDK

        if _LAZY_GLOBAL_SDK is None:
            _LAZY_GLOBAL_SDK = ServiceSDK()

        return f(*args, _sdk=_LAZY_GLOBAL_SDK, **kwargs)

    return wrapper


@_inject_global_sdk
def deploy(
    config: ServiceConfig,
    *,
    in_place: bool = False,
    canary_percent: Optional[int] = None,
    max_surge_percent: Optional[int] = None,
    _sdk: ServiceSDK,
):
    return _sdk.deploy(
        config,
        in_place=in_place,
        canary_percent=canary_percent,
        max_surge_percent=max_surge_percent,
    )


@_inject_global_sdk
def status(name: str, *, _sdk: ServiceSDK) -> ServiceStatus:
    return _sdk.status(name)



import copy
import os
from typing import Any, Dict, Generator, Optional
from unittest.mock import patch

import click
from click.testing import CliRunner
import pytest

from anyscale.commands.service_commands import deploy
from anyscale.service.models import ServiceConfig, ServiceStatus


MINIMAL_CONFIG_PATH = os.path.join(
    os.path.dirname(__file__), "test_config_files", "minimal.yaml",
)
FULL_CONFIG_PATH = os.path.join(
    os.path.dirname(__file__), "test_config_files", "full.yaml",
)
MULTI_LINE_REQUIREMENTS_PATH = os.path.join(
    os.path.dirname(__file__), "test_requirements_files", "multi_line.txt",
)


class FakeServiceSDK:
    def __init__(self):
        self._deployed_config: Optional[ServiceConfig] = None
        self._deployed_kwargs: Dict[str, Any] = {}

    @property
    def deployed_config(self) -> Optional[ServiceConfig]:
        return self._deployed_config

    @property
    def deployed_kwargs(self) -> Dict[str, Any]:
        return copy.deepcopy(self._deployed_kwargs)

    def deploy(self, config: ServiceConfig, **kwargs):
        assert isinstance(config, ServiceConfig)
        self._deployed_config = config
        self._deployed_kwargs = kwargs

    def status(self) -> ServiceStatus:
        return ServiceStatus()


@pytest.fixture()
def fake_service_sdk() -> Generator[FakeServiceSDK, None, None]:
    fake_service_sdk = FakeServiceSDK()
    with patch(
        "anyscale.service.commands._LAZY_GLOBAL_SDK", new=fake_service_sdk,
    ):
        yield fake_service_sdk


def _assert_error_message(result: click.testing.Result, *, message: str):
    assert result.exit_code != 0
    assert message in result.stdout


class TestDeploy:
    def test_deploy_from_import_path(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(import_path="main:app")
        assert (
            fake_service_sdk.deployed_config.ray_serve_config
            == expected_config.ray_serve_config
        )

    def test_deploy_from_import_path_with_args(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "arg1=val1", "arg2=val2"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            import_path="main:app", arguments={"arg1": "val1", "arg2": "val2"}
        )
        assert (
            fake_service_sdk.deployed_config.ray_serve_config
            == expected_config.ray_serve_config
        )

    def test_deploy_from_import_path_with_bad_arg(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "bad_arg"])
        _assert_error_message(result, message="Invalid application argument 'bad_arg'")

    def test_deploy_from_file(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, [MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(import_path="main:app")
        assert (
            fake_service_sdk.deployed_config.ray_serve_config
            == expected_config.ray_serve_config
        )

    def test_deploy_from_file_with_args(self, fake_service_sdk):
        runner = CliRunner()
        os.path.join(
            os.path.dirname(__file__), "test_config_files", "minimal.yaml",
        )
        result = runner.invoke(deploy, [MINIMAL_CONFIG_PATH, "arg1=val1"])
        _assert_error_message(
            result,
            message="Application arguments can't be passed when deploying from a config file.",
        )

    @pytest.mark.parametrize("flag", ["--in-place", "-i"])
    def test_deploy_in_place(self, fake_service_sdk, flag: str):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", flag])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(import_path="main:app")
        assert (
            fake_service_sdk.deployed_config.ray_serve_config
            == expected_config.ray_serve_config
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": None,
            "in_place": True,
            "max_surge_percent": None,
        }

    def test_deploy_canary_percent(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "--canary-percent", "50"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(import_path="main:app")
        assert (
            fake_service_sdk.deployed_config.ray_serve_config
            == expected_config.ray_serve_config
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": 50,
            "in_place": False,
            "max_surge_percent": None,
        }

    def test_deploy_max_surge_percent(self, fake_service_sdk):
        runner = CliRunner()
        result = runner.invoke(deploy, ["main:app", "--max-surge-percent", "50"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(import_path="main:app")
        assert (
            fake_service_sdk.deployed_config.ray_serve_config
            == expected_config.ray_serve_config
        )
        assert fake_service_sdk.deployed_kwargs == {
            "canary_percent": None,
            "in_place": False,
            "max_surge_percent": 50,
        }

    def test_deploy_from_file_override_options(self, fake_service_sdk):
        runner = CliRunner()

        # No overrides, should match the config in the file.
        result = runner.invoke(deploy, [FULL_CONFIG_PATH])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            import_path="main:app",
            name="test-name",
            image="test-image",
            working_dir="test-working-dir",
            requirements=["pip-install-test"],
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override name.
        result = runner.invoke(deploy, [FULL_CONFIG_PATH, "--name", "override-name"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            import_path="main:app",
            name="override-name",
            image="test-image",
            working_dir="test-working-dir",
            requirements=["pip-install-test"],
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override image.
        result = runner.invoke(deploy, [FULL_CONFIG_PATH, "--image", "override-image"])
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            import_path="main:app",
            name="test-name",
            image="override-image",
            working_dir="test-working-dir",
            requirements=["pip-install-test"],
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override working_dir.
        result = runner.invoke(
            deploy, [FULL_CONFIG_PATH, "--working-dir", "override-working-dir"]
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            import_path="main:app",
            name="test-name",
            image="test-image",
            working_dir="override-working-dir",
            requirements=["pip-install-test"],
        )
        assert fake_service_sdk.deployed_config == expected_config

        # Override requirements.
        result = runner.invoke(
            deploy, [FULL_CONFIG_PATH, "--requirements", MULTI_LINE_REQUIREMENTS_PATH]
        )
        assert result.exit_code == 0
        assert fake_service_sdk.deployed_config is not None

        expected_config = ServiceConfig(
            import_path="main:app",
            name="test-name",
            image="test-image",
            working_dir="test-working-dir",
            requirements=["pip-install-test", "torch==1.10.1"],
        )
        assert fake_service_sdk.deployed_config == expected_config



from dataclasses import dataclass
import os
from typing import Any, Generator, Tuple
from unittest.mock import Mock, patch

import pytest

from anyscale.client.openapi_client.configuration import Configuration
from anyscale.client.openapi_client.models import (
    CloudDataBucketFileType,
    CloudDataBucketPresignedUploadInfo,
    CloudDataBucketPresignedUploadRequest,
    Project,
)
from anyscale.service._private.anyscale_client import RealAnyscaleClient


def _get_test_working_dir_path(name: str) -> str:
    return os.path.join(os.path.dirname(__file__), "test_working_dirs", name)


BASIC_WORKING_DIR = _get_test_working_dir_path("basic")
NESTED_WORKING_DIR = _get_test_working_dir_path("nested")
SYMLINK_WORKING_DIR = _get_test_working_dir_path("symlink_to_basic")
TEST_WORKING_DIRS = [BASIC_WORKING_DIR, NESTED_WORKING_DIR, SYMLINK_WORKING_DIR]

OPENAPI_NO_VALIDATION = Configuration()
OPENAPI_NO_VALIDATION.client_side_validation = False


class FakeServiceController:
    pass


@dataclass
class FakeClientResult:
    result: Any


class FakeInternalAPIClient:
    FAKE_FILE_URI = "s3://some-bucket/{file_name}"
    FAKE_UPLOAD_URL = "http://some-domain.com/upload-magic-file/{file_name}"
    DEFAULT_PROJECT_ID = "fake-default-project-id"

    def __init__(self):
        self._num_get_project_calls: int = 0

    @property
    def num_get_project_calls(self) -> int:
        return self._num_get_project_calls

    def get_default_project_api_v2_projects_default_project_get(
        self,
    ) -> FakeClientResult:
        self._num_get_project_calls += 1
        return FakeClientResult(
            result=Project(
                id=self.DEFAULT_PROJECT_ID,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )

    def generate_cloud_data_bucket_presigned_upload_url_api_v2_clouds_cloud_id_generate_cloud_data_bucket_presigned_upload_url_post(
        self, cloud_id: str, request: CloudDataBucketPresignedUploadRequest
    ) -> FakeClientResult:
        assert request.file_type == CloudDataBucketFileType.RUNTIME_ENV_PACKAGES
        assert isinstance(request.file_name, str)
        return FakeClientResult(
            result=CloudDataBucketPresignedUploadInfo(
                upload_url=self.FAKE_UPLOAD_URL.format(file_name=request.file_name),
                file_uri=self.FAKE_FILE_URI.format(file_name=request.file_name),
            ),
        )


@pytest.fixture()
def setup_anyscale_client() -> Tuple[RealAnyscaleClient, FakeInternalAPIClient]:
    fake_internal_client = FakeInternalAPIClient()
    return (
        RealAnyscaleClient(
            controller=FakeServiceController(),
            internal_api_client=fake_internal_client,
        ),
        fake_internal_client,
    )


@pytest.fixture()
def mock_requests_put() -> Generator[Mock, None, None]:
    with patch("requests.put") as mock_requests_put:
        yield mock_requests_put


def test_get_default_project(
    setup_anyscale_client: Tuple[RealAnyscaleClient, FakeInternalAPIClient]
):
    anyscale_client, fake_internal_client = setup_anyscale_client

    # The project ID should be cached so we only make one API call.
    for _ in range(100):
        assert (
            anyscale_client.get_default_project_id()
            == FakeInternalAPIClient.DEFAULT_PROJECT_ID
        )
        assert fake_internal_client.num_get_project_calls == 1


class TestUploadLocalDirToCloudStorage:
    @pytest.mark.parametrize("working_dir", TEST_WORKING_DIRS)
    def test_basic(
        self,
        setup_anyscale_client: Tuple[RealAnyscaleClient, FakeInternalAPIClient],
        mock_requests_put: Mock,
        working_dir: str,
    ):
        anyscale_client, _ = setup_anyscale_client
        uri = anyscale_client.upload_local_dir_to_cloud_storage(
            working_dir, cloud_id="test-cloud-id",
        )
        assert isinstance(uri, str) and len(uri) > 0
        mock_requests_put.assert_called_once()
        mock_requests_put.reset_mock()

    def test_missing_dir(
        self,
        setup_anyscale_client: Tuple[RealAnyscaleClient, FakeInternalAPIClient],
        mock_requests_put: Mock,
    ):
        anyscale_client, _ = setup_anyscale_client
        with pytest.raises(
            RuntimeError, match="Path 'does_not_exist' is not a valid directory."
        ):
            anyscale_client.upload_local_dir_to_cloud_storage(
                "does_not_exist", cloud_id="test-cloud-id",
            )

    def test_uri_content_addressed(
        self,
        setup_anyscale_client: Tuple[RealAnyscaleClient, FakeInternalAPIClient],
        mock_requests_put: Mock,
    ):
        anyscale_client, _ = setup_anyscale_client

        # Uploading the same directory contents should result in the same content-addressed URI.
        uri1 = anyscale_client.upload_local_dir_to_cloud_storage(
            BASIC_WORKING_DIR, cloud_id="test-cloud-id",
        )
        uri2 = anyscale_client.upload_local_dir_to_cloud_storage(
            BASIC_WORKING_DIR, cloud_id="test-cloud-id",
        )
        assert uri1 == uri2

        # Uploading a different directory should not result in the same content-addressed URI.
        uri3 = anyscale_client.upload_local_dir_to_cloud_storage(
            NESTED_WORKING_DIR, cloud_id="test-cloud-id",
        )
        assert uri1 != uri3 and uri2 != uri3



from typing import Any, Dict, Optional

from anyscale.service._private.anyscale_client import (
    AnyscaleClientWrapper,
    RealAnyscaleClient,
)
from anyscale.service.models import ServiceConfig, ServiceStatus


class ServiceSDK:
    def __init__(self, *, client: Optional[AnyscaleClientWrapper] = None):
        self._client = client or RealAnyscaleClient()

    def deploy(
        self,
        config: ServiceConfig,
        *,
        in_place: bool = False,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
    ):
        if not isinstance(in_place, bool):
            raise TypeError("in_place must be a bool.")

        if canary_percent is not None:
            if not isinstance(canary_percent, int):
                raise TypeError("canary_percent must be an int.")
            if canary_percent < 0 or canary_percent > 100:
                raise ValueError("canary_percent must be between 0 and 100.")

        if max_surge_percent is not None:
            if not isinstance(max_surge_percent, int):
                raise TypeError("max_surge_percent must be an int.")

            if max_surge_percent < 0 or max_surge_percent > 100:
                raise ValueError("max_surge_percent must be between 0 and 100.")

        config = config.with_local_dirs_uploaded(self._client)
        config_dict: Dict[str, Any] = {
            "ray_serve_config": config.ray_serve_config,
            "rollout_strategy": "IN_PLACE" if in_place else "ROLLOUT",
            "project_id": self._client.get_default_project_id(),
        }
        if config.name is not None:
            config_dict["name"] = config.name
        if config.image is not None:
            config_dict["cluster_env"] = config.image
        if canary_percent is not None:
            config_dict["canary_percent"] = canary_percent
        if max_surge_percent is not None:
            config_dict["max_surge_percent"] = max_surge_percent

        self._client.roll_out_service(config_dict)

    def status(self, name: str) -> ServiceStatus:  # noqa: ARG002
        return ServiceStatus()



from typing import Tuple

import pytest

from anyscale.service._private.anyscale_client import FakeAnyscaleClient
from anyscale.service._private.sdk import ServiceSDK
from anyscale.service.models import ServiceConfig


@pytest.fixture()
def sdk_with_fake_client() -> Tuple[ServiceSDK, FakeAnyscaleClient]:
    fake_client = FakeAnyscaleClient()
    return ServiceSDK(client=fake_client), fake_client


class TestDeploy:
    def test_validation(
        self, sdk_with_fake_client: Tuple[ServiceSDK, FakeAnyscaleClient]
    ):
        sdk, client = sdk_with_fake_client

        config = ServiceConfig(import_path="main:app")

        with pytest.raises(TypeError):
            sdk.deploy(config, name=123)

        with pytest.raises(TypeError):
            sdk.deploy(config, in_place="in_place")

        with pytest.raises(TypeError):
            sdk.deploy(config, canary_percent="10")

        with pytest.raises(ValueError):
            sdk.deploy(config, canary_percent=-1)

        with pytest.raises(ValueError):
            sdk.deploy(config, canary_percent=101)

        with pytest.raises(TypeError):
            sdk.deploy(config, max_surge_percent="10")

        with pytest.raises(ValueError):
            sdk.deploy(config, max_surge_percent=-1)

        with pytest.raises(ValueError):
            sdk.deploy(config, max_surge_percent=101)

    def test_basic(self, sdk_with_fake_client: Tuple[ServiceSDK, FakeAnyscaleClient]):
        sdk, client = sdk_with_fake_client

        config = ServiceConfig(import_path="main:app")
        sdk.deploy(config)
        assert client.rolled_out_service == {
            "project_id": client.DEFAULT_PROJECT_ID,
            "ray_serve_config": config.ray_serve_config,
            "rollout_strategy": "ROLLOUT",
        }

    def test_in_place(
        self, sdk_with_fake_client: Tuple[ServiceSDK, FakeAnyscaleClient]
    ):
        sdk, client = sdk_with_fake_client

        config = ServiceConfig(import_path="main:app")
        sdk.deploy(config, in_place=True)
        assert client.rolled_out_service == {
            "project_id": client.DEFAULT_PROJECT_ID,
            "ray_serve_config": config.ray_serve_config,
            "rollout_strategy": "IN_PLACE",
        }

    def test_canary_percent(
        self, sdk_with_fake_client: Tuple[ServiceSDK, FakeAnyscaleClient]
    ):
        sdk, client = sdk_with_fake_client

        config = ServiceConfig(import_path="main:app")
        sdk.deploy(config, canary_percent=50)
        assert client.rolled_out_service == {
            "project_id": client.DEFAULT_PROJECT_ID,
            "ray_serve_config": config.ray_serve_config,
            "rollout_strategy": "ROLLOUT",
            "canary_percent": 50,
        }

    def test_max_surge_percent(
        self, sdk_with_fake_client: Tuple[ServiceSDK, FakeAnyscaleClient]
    ):
        sdk, client = sdk_with_fake_client

        config = ServiceConfig(import_path="main:app")
        sdk.deploy(config, max_surge_percent=50)
        assert client.rolled_out_service == {
            "project_id": client.DEFAULT_PROJECT_ID,
            "ray_serve_config": config.ray_serve_config,
            "rollout_strategy": "ROLLOUT",
            "max_surge_percent": 50,
        }

    def test_upload_local_dirs(
        self, sdk_with_fake_client: Tuple[ServiceSDK, FakeAnyscaleClient]
    ):
        sdk, client = sdk_with_fake_client

        config = ServiceConfig(import_path="main:app", working_dir=".")
        sdk.deploy(config)
        [application] = client.rolled_out_service["ray_serve_config"]["applications"]
        assert application["import_path"] == "main:app"
        assert application["runtime_env"]["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )



from dataclasses import dataclass
import os
import re
from typing import List, Optional

import pytest

from anyscale.service._private.anyscale_client import FakeAnyscaleClient
from anyscale.service.models import ServiceConfig


@dataclass
class ConfigFile:
    name: str
    expected_config: Optional[ServiceConfig] = None
    expected_error: Optional[str] = None

    def get_path(self) -> str:
        return os.path.join(os.path.dirname(__file__), "test_config_files", self.name)


TEST_CONFIG_FILES = [
    ConfigFile("minimal.yaml", expected_config=ServiceConfig(import_path="main:app")),
    ConfigFile(
        "full.yaml",
        expected_config=ServiceConfig(
            import_path="main:app",
            name="test-name",
            image="test-image",
            working_dir="test-working-dir",
            requirements=["pip-install-test"],
        ),
    ),
    ConfigFile(
        "arguments.yaml",
        expected_config=ServiceConfig(
            import_path="main:app", arguments={"abc": "def", "nested": {"key": "val"}}
        ),
    ),
    ConfigFile(
        "unrecognized_option.yaml",
        expected_error=re.escape("Unrecognized options: ['bad_option']."),
    ),
    ConfigFile(
        "import_path_and_ray_serve_config.yaml",
        expected_error=re.escape(
            "import_path and arguments cannot be provided with ray_serve_config."
        ),
    ),
    ConfigFile(
        "ray_serve_config.yaml",
        expected_config=ServiceConfig(
            ray_serve_config={
                "applications": [
                    {
                        "import_path": "main:app",
                        "runtime_env": {"env_vars": {"abc": "def"}},
                        "arguments": {"abc": "def", "nested": {"key": "val"}},
                    }
                ],
            },
        ),
    ),
]


@dataclass
class RequirementsFile:
    name: str
    expected_pip_list: Optional[List[str]]

    def get_path(self) -> str:
        return os.path.join(
            os.path.dirname(__file__), "test_requirements_files", self.name,
        )


TEST_REQUIREMENTS_FILES = [
    RequirementsFile("does_not_exist.txt", None),
    RequirementsFile("empty.txt", []),
    RequirementsFile("single_line.txt", ["pip-install-test"]),
    RequirementsFile("multi_line.txt", ["pip-install-test", "torch==1.10.1"]),
    RequirementsFile(
        "multi_line_with_whitespace.txt",
        ["pip-install-test", "torch==1.10.1", "something-else"],
    ),
    RequirementsFile("comments.txt", ["pip-install-test", "torch==1.10.1"]),
]


class TestServiceConfig:
    def test_no_ray_serve_config_fields(self):
        with pytest.raises(
            ValueError, match="Either import_path or ray_serve_config must be provided."
        ):
            ServiceConfig(
                import_path=None, ray_serve_config=None,
            )

    def test_reject_import_path_and_config(self):
        with pytest.raises(
            ValueError,
            match="import_path and arguments cannot be provided with ray_serve_config.",
        ):
            ServiceConfig(
                import_path="main:app", ray_serve_config={"test": "123"},
            )

    def test_reject_arguments_and_config(self):
        with pytest.raises(
            ValueError,
            match="import_path and arguments cannot be provided with ray_serve_config.",
        ):
            ServiceConfig(
                ray_serve_config={"test": "123"}, arguments={"abc": "def"},
            )

    def test_import_path(self):
        config = ServiceConfig(import_path="main:app")
        assert config.ray_serve_config == {
            "applications": [{"import_path": "main:app",},],
        }

    def test_import_path_with_arguments(self):
        config = ServiceConfig(import_path="main:app", arguments={"hello": "world"})
        assert config.ray_serve_config == {
            "applications": [
                {"import_path": "main:app", "arguments": {"hello": "world"},},
            ],
        }

    def test_import_path_with_nested_arguments(self):
        config = ServiceConfig(
            import_path="main:app",
            arguments={"hello": "world", "nested": {"hi": "there"}, "list": [1]},
        )
        assert config.ray_serve_config == {
            "applications": [
                {
                    "import_path": "main:app",
                    "arguments": {
                        "hello": "world",
                        "nested": {"hi": "there"},
                        "list": [1],
                    },
                },
            ],
        }

    def test_import_path_with_bad_arguments(self):
        err_msg = "Application arguments must be a dictionary."
        with pytest.raises(TypeError, match=err_msg):
            ServiceConfig(
                import_path="main:app", arguments=["a", "b", "c"],
            )

    def test_passed_ray_serve_config(self):
        ray_serve_config = {
            "applications": [{"import_path": "main:app",},],
        }
        config = ServiceConfig(ray_serve_config=ray_serve_config)
        assert config.ray_serve_config == ray_serve_config

    def test_name(self):
        config = ServiceConfig(import_path="main:app")
        assert config.name is None

        config = ServiceConfig(import_path="main:app", name="my-custom-name")
        assert config.name == "my-custom-name"

        with pytest.raises(TypeError, match="name must be a string"):
            ServiceConfig(import_path="main:app", name=123)

    def test_image(self):
        config = ServiceConfig(import_path="main:app")
        assert config.image is None

        config = ServiceConfig(import_path="main:app", image="my-custom-image:1")
        assert config.image == "my-custom-image:1"

        with pytest.raises(TypeError, match="image must be a string"):
            ServiceConfig(import_path="main:app", image=123)

    def test_options(self):
        config = ServiceConfig(import_path="main:app", arguments={"hello": "world"})

        options = {
            "name": "test-name",
            "image": "test-image",
            "requirements": ["pip-install-test"],
            "working_dir": ".",
        }

        # Test setting fields one at a time.
        for option, val in options.items():
            assert config.options(**{option: val}) == ServiceConfig(
                import_path="main:app", arguments={"hello": "world"}, **{option: val}
            )

            assert config.options(**{option: val}) == ServiceConfig(
                import_path="main:app", arguments={"hello": "world"}, **{option: val}
            )

        # Test setting fields all at once.
        assert config.options(**options) == ServiceConfig(
            import_path="main:app", arguments={"hello": "world"}, **options
        )

    @pytest.mark.parametrize("working_dir", [None, ".", "s3://path.zip"])
    def test_override_working_dir_import_path(self, working_dir: Optional[str]):
        config = ServiceConfig(
            import_path="main:app", arguments={"foo": "bar"}, working_dir=working_dir,
        )

        if working_dir is None:
            assert "runtime_env" not in config.ray_serve_config["applications"][0]
        else:
            config.ray_serve_config["applications"][0]["runtime_env"] = {
                "working_dir": working_dir,
            }

    @pytest.mark.parametrize("working_dir", [None, ".", "s3://path.zip"])
    def test_override_working_dir_ray_serve_config(self, working_dir: Optional[str]):
        ray_serve_config = {
            "applications": [
                {"name": "no_runtime_env", "import_path": "main:app",},
                {
                    "name": "empty_runtime_env",
                    "import_path": "main:app",
                    "runtime_env": {},
                },
                {
                    "name": "has_runtime_env",
                    "import_path": "main:app",
                    "runtime_env": {
                        "working_dir": "s3://somewhere.zip",
                        "env_vars": {"abc": "123"},
                    },
                },
            ],
        }

        config = ServiceConfig(
            ray_serve_config=ray_serve_config, working_dir=working_dir,
        )

        if working_dir is None:
            assert config.ray_serve_config == ray_serve_config
        else:
            assert len(config.ray_serve_config["applications"]) == 3
            assert config.ray_serve_config["applications"][0] == {
                "name": "no_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"working_dir": working_dir,},
            }

            assert config.ray_serve_config["applications"][1] == {
                "name": "empty_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"working_dir": working_dir,},
            }
            assert config.ray_serve_config["applications"][2] == {
                "name": "has_runtime_env",
                "import_path": "main:app",
                "runtime_env": {
                    "working_dir": working_dir,
                    "env_vars": {"abc": "123"},
                },
            }

    @pytest.mark.parametrize("requirements", [None, *TEST_REQUIREMENTS_FILES])
    def test_override_requirements_file(self, requirements: Optional[RequirementsFile]):
        ray_serve_config = {
            "applications": [
                {"name": "no_runtime_env", "import_path": "main:app",},
                {
                    "name": "empty_runtime_env",
                    "import_path": "main:app",
                    "runtime_env": {},
                },
                {
                    "name": "has_runtime_env",
                    "import_path": "main:app",
                    "runtime_env": {
                        "env_vars": {"abc": "123"},
                        "working_dir": "s3://somewhere.zip",
                    },
                },
            ],
        }

        if requirements is not None and requirements.expected_pip_list is None:
            with pytest.raises(ValueError):
                ServiceConfig(
                    ray_serve_config=ray_serve_config,
                    requirements=requirements.get_path(),
                )

            return

        config = ServiceConfig(
            ray_serve_config=ray_serve_config,
            requirements=requirements.get_path() if requirements else None,
        )
        if requirements is None:
            assert config.ray_serve_config == ray_serve_config
        else:
            assert len(config.ray_serve_config["applications"]) == 3
            assert config.ray_serve_config["applications"][0] == {
                "name": "no_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"pip": requirements.expected_pip_list,},
            }

            assert config.ray_serve_config["applications"][1] == {
                "name": "empty_runtime_env",
                "import_path": "main:app",
                "runtime_env": {"pip": requirements.expected_pip_list,},
            }
            assert config.ray_serve_config["applications"][2] == {
                "name": "has_runtime_env",
                "import_path": "main:app",
                "runtime_env": {
                    "env_vars": {"abc": "123"},
                    "working_dir": "s3://somewhere.zip",
                    "pip": requirements.expected_pip_list,
                },
            }

    @pytest.mark.parametrize("config_file", TEST_CONFIG_FILES)
    def test_from_config_file(self, config_file: ConfigFile):
        if config_file.expected_error is not None:
            with pytest.raises(Exception, match=config_file.expected_error):
                ServiceConfig.from_yaml(config_file.get_path())

            return

        assert config_file.expected_config == ServiceConfig.from_yaml(
            config_file.get_path()
        )

    def test_upload_local_working_dir(self):
        client = FakeAnyscaleClient()
        config = ServiceConfig(import_path="main:app", working_dir=".",)
        new_config = config.with_local_dirs_uploaded(client)

        # The original config should not be modified.
        assert config != new_config

        applications = new_config.ray_serve_config["applications"]
        assert len(applications) == 1
        assert applications[0]["runtime_env"]["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )

    def test_upload_with_no_local_dirs(self):
        """Configs should be left unchanged if there are no local dirs."""
        client = FakeAnyscaleClient()

        basic_config = ServiceConfig(import_path="main:app",)
        assert basic_config == basic_config.with_local_dirs_uploaded(client)

        config_with_requirements = ServiceConfig(
            import_path="main:app", requirements=TEST_REQUIREMENTS_FILES[2],
        )
        assert (
            config_with_requirements
            == config_with_requirements.with_local_dirs_uploaded(client)
        )

        complex_config = ServiceConfig(
            ray_serve_config={
                "applications": [
                    {"name": "app1", "import_path": "main:app",},
                    {
                        "name": "app2",
                        "import_path": "main:app",
                        "runtime_env": {"env_vars": {"foo": "bar",},},
                    },
                ],
            },
        )
        assert complex_config == complex_config.with_local_dirs_uploaded(client)

    def test_no_upload_remote_working_dir(self):
        client = FakeAnyscaleClient()
        config = ServiceConfig(
            import_path="main:app", working_dir="s3://some-remote-uri.zip",
        )
        config = config.with_local_dirs_uploaded(client)

        applications = config.ray_serve_config["applications"]
        assert len(applications) == 1
        assert (
            applications[0]["runtime_env"]["working_dir"] == "s3://some-remote-uri.zip"
        )

    def test_upload_local_py_modules(self):
        client = FakeAnyscaleClient()
        config = ServiceConfig(
            ray_serve_config={
                "applications": [
                    {
                        "import_path": "main:app",
                        "runtime_env": {
                            "py_modules": [
                                # Should be left alone.
                                "s3://some-remote-uri.zip",
                                # Should be uploaded.
                                "local-path",
                            ],
                        },
                    },
                ],
            },
        )
        config = config.with_local_dirs_uploaded(client)
        applications = config.ray_serve_config["applications"]
        assert len(applications) == 1
        assert (
            applications[0]["runtime_env"]["py_modules"][0]
            == "s3://some-remote-uri.zip"
        )
        assert applications[0]["runtime_env"]["py_modules"][1].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )

    def test_upload_caching(self):
        """The same directory should only by uploaded once."""
        client = FakeAnyscaleClient()
        config = ServiceConfig(import_path="main:app", working_dir=".",)
        config = ServiceConfig(
            ray_serve_config={
                "applications": [
                    {
                        "name": "app1",
                        "import_path": "main:app",
                        "runtime_env": {"working_dir": ".",},
                    },
                    {
                        "name": "app2",
                        "import_path": "main:app",
                        "runtime_env": {
                            "working_dir": ".",
                            "py_modules": [".", "other-dir",],
                        },
                    },
                ],
            },
        )
        config = config.with_local_dirs_uploaded(client)

        applications = config.ray_serve_config["applications"]
        assert len(applications) == 2
        assert applications[0]["runtime_env"]["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )
        common_uri = applications[0]["runtime_env"]["working_dir"]

        assert applications[1]["runtime_env"]["working_dir"] == common_uri
        assert applications[1]["runtime_env"]["py_modules"][0] == common_uri
        assert applications[1]["runtime_env"]["py_modules"][1] != common_uri
        assert applications[1]["runtime_env"]["py_modules"][1].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(
                cloud_id=FakeAnyscaleClient.DEFAULT_CLOUD_ID
            )
        )
